from __future__ import annotations

import re
import string
import subprocess
from collections import ChainMap
from math import hypot
from pathlib import Path
from typing import Optional

import plotly.express as px
import workflows.recipe
from pydantic import BaseModel, Field, ValidationError, validator
from workflows.services.common_service import CommonService

from relion.zocalo.spa_relion_service_options import RelionServiceOptions


class MotionCorrParameters(BaseModel):
    movie: str = Field(..., min_length=1)
    mrc_out: str = Field(..., min_length=1)
    experiment_type: str
    pix_size: float
    fm_dose: float
    patch_size: dict = {"x": 5, "y": 5}
    gpu: int = 0
    gain_ref: str = None
    rot_gain: int = None
    flip_gain: int = None
    dark: str = None
    use_gpus: int = None
    sum_range: Optional[dict] = None
    iter: int = None
    tol: float = None
    throw: int = None
    trunc: int = None
    fm_ref: int = 0
    kv: int = None
    fm_int_file: str = None
    mag: Optional[dict] = None
    ft_bin: float = None
    serial: int = None
    in_suffix: str = None
    eer_sampling: int = None
    out_stack: int = None
    bft: Optional[dict] = None
    group: int = None
    defect_file: str = None
    arc_dir: str = None
    in_fm_motion: int = None
    split_sum: int = None
    dose_motionstats_cutoff: float = 4.0
    movie_id: int
    mc_uuid: int
    picker_uuid: int
    relion_options: Optional[RelionServiceOptions] = None
    ctf: dict = {}

    @validator("experiment_type")
    def is_spa_or_tomo(cls, experiment):
        if experiment not in ["spa", "tomography"]:
            raise ValueError("Specify an experiment type of spa or tomography.")
        return experiment

    class Config:
        ignore_extra = True


class ChainMapWithReplacement(ChainMap):
    def __init__(self, *maps, substitutions=None) -> None:
        super().__init__(*maps)
        self._substitutions = substitutions

    def __getitem__(self, k):
        v = super().__getitem__(k)
        if self._substitutions and isinstance(v, str) and "$" in v:
            template = string.Template(v)
            return template.substitute(**self._substitutions)
        return v


class MotionCorr(CommonService):
    """
    A service for motion correcting cryoEM movies using MotionCor2
    """

    # Human readable service name
    _service_name = "MotionCorr"

    # Logger name
    _logger_name = "relion.zocalo.motioncorr"

    # Job name
    job_type = "relion.motioncorr.motioncor2"

    # Values to extract for ISPyB
    x_shift_list = []
    y_shift_list = []

    def initializing(self):
        """Subscribe to a queue. Received messages must be acknowledged."""
        self.log.info("Motion correction service starting")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "motioncorr",
            self.motion_correction,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    def parse_mc_output(self, mc_stdout: str):
        """
        Read the output logs of MotionCorr to determine
        the movement of each frame
        """
        frames_line = False
        for line in mc_stdout.split("\n"):
            # Frame reading in MotionCorr 1.4.0
            if line.startswith("...... Frame"):
                line_split = line.split()
                self.x_shift_list.append(float(line_split[-2]))
                self.y_shift_list.append(float(line_split[-1]))

            # Alternative frame reading for MotionCorr 1.6.3
            if not line:
                frames_line = False
            if frames_line:
                line_split = line.split()
                self.x_shift_list.append(float(line_split[1]))
                self.y_shift_list.append(float(line_split[2]))
            if "x Shift" in line:
                frames_line = True

    def motioncor2(self, command, mrc_out):
        """Run the MotionCor2 command"""
        result = subprocess.run(command, capture_output=True)
        self.parse_mc_output(result.stdout.decode("utf8", "replace"))
        return result

    def motion_correction(self, rw, header: dict, message: dict):
        class MockRW:
            def dummy(self, *args, **kwargs):
                pass

        if not rw:
            if (
                not isinstance(message, dict)
                or not message.get("parameters")
                or not message.get("content")
            ):
                self.log.error("Rejected invalid simple message")
                self._transport.nack(header)
                return

            # Create a wrapper-like object that can be passed to functions
            # as if a recipe wrapper was present.
            rw = MockRW()
            rw.transport = self._transport
            rw.recipe_step = {"parameters": message["parameters"], "output": None}
            rw.environment = {"has_recipe_wrapper": False}
            rw.set_default_channel = rw.dummy
            rw.send = rw.dummy
            message = message["content"]

        command = ["MotionCor2"]

        parameter_map = ChainMapWithReplacement(
            message if isinstance(message, dict) else {},
            rw.recipe_step["parameters"],
            substitutions=rw.environment,
        )

        try:
            if isinstance(message, dict):
                mc_params = MotionCorrParameters(**{**dict(parameter_map), **message})
            else:
                mc_params = MotionCorrParameters(**{**dict(parameter_map)})
        except (ValidationError, TypeError) as e:
            self.log.warning(
                f"Motion correction parameter validation failed for message: {message} "
                f"and recipe parameters: {rw.recipe_step.get('parameters', {})} "
                f"with exception: {e}"
            )
            rw.transport.nack(header)
            return

        mc_flags = {
            "mrc_out": "-OutMrc",
            "patch_size": "-Patch",
            "pix_size": "-PixSize",
            "gain_ref": "-Gain",
            "rot_gain": "-RotGain",
            "flip_gain": "-FlipGain",
            "dark": "-Dark",
            "gpu": "-Gpu",
            "use_gpus": "-UseGpus",
            "sum_range": "-SumRange",
            "iter": "-Iter",
            "tol": "-Tol",
            "throw": "-Throw",
            "trunc": "-Trunc",
            "fm_ref": "-FmRef",
            "kv": "-Kv",
            "fm_dose": "-FmDose",
            "mag": "-Mag",
            "ft_bin": "-FtBin",
            "serial": "-Serial",
            "in_suffix": "-InSuffix",
            "eer_sampling": "-EerSampling",
            "out_stack": "-OutStack",
            "bft": "-Bft",
            "group": "-Group",
            "defect_file": "-DefectFile",
            "arc_dir": "-ArcDir",
            "in_fm_motion": "-InFmMotion",
            "split_sum": "-SplitSum",
        }

        # Determine the input and output files
        if not Path(mc_params.mrc_out).parent.exists():
            Path(mc_params.mrc_out).parent.mkdir(parents=True)
        if mc_params.movie.endswith(".mrc"):
            input_flag = "-InMrc"
        elif mc_params.movie.endswith((".tif", ".tiff")):
            input_flag = "-InTiff"
        elif mc_params.movie.endswith(".eer"):
            input_flag = "-InEer"
            mc_flags["fm_int_file"] = "-FmIntFile"
        else:
            self.log.error(f"No input flag found for movie {mc_params.movie}")
            input_flag = None
            rw.transport.nack(header)
        command.extend([input_flag, mc_params.movie])

        # Create the motion correction command
        for k, v in mc_params.dict().items():
            if (v is not None) and (k in mc_flags):
                if type(v) is dict:
                    command.extend((mc_flags[k], " ".join(str(_) for _ in v.values())))
                else:
                    command.extend((mc_flags[k], str(v)))

        self.log.info(f"Input: {mc_params.movie} Output: {mc_params.mrc_out}")

        # Run motion correction and confirm it ran successfully
        result = self.motioncor2(command, mc_params.mrc_out)
        if result.returncode:
            self.log.error(
                f"Motion correction of {mc_params.movie} "
                f"failed with exitcode {result.returncode}:\n"
                + result.stderr.decode("utf8", "replace")
            )
            if mc_params.experiment_type == "spa":
                # On spa failure send the outputs to the node creator
                node_creator_parameters = {
                    "job_type": self.job_type,
                    "input_file": mc_params.mrc_out,
                    "output_file": mc_params.mrc_out,
                    "relion_options": dict(mc_params.relion_options),
                    "command": " ".join(command),
                    "stdout": result.stdout.decode("utf8", "replace"),
                    "stderr": result.stderr.decode("utf8", "replace"),
                    "success": False,
                }
                if isinstance(rw, MockRW):
                    rw.transport.send(
                        destination="node_creator",
                        message={
                            "parameters": node_creator_parameters,
                            "content": "dummy",
                        },
                    )
                else:
                    rw.send_to("node_creator", node_creator_parameters)
            rw.transport.nack(header)
            return

        # Extract results for ispyb
        total_motion = 0
        early_motion = 0
        late_motion = 0
        cutoff_frame = round(mc_params.dose_motionstats_cutoff / mc_params.fm_dose)
        for i in range(1, len(self.x_shift_list)):
            total_motion += hypot(
                self.x_shift_list[i] - self.x_shift_list[i - 1],
                self.y_shift_list[i] - self.y_shift_list[i - 1],
            )
            if i < cutoff_frame:
                early_motion += hypot(
                    self.x_shift_list[i] - self.x_shift_list[i - 1],
                    self.y_shift_list[i] - self.y_shift_list[i - 1],
                )
            else:
                late_motion += hypot(
                    self.x_shift_list[i] - self.x_shift_list[i - 1],
                    self.y_shift_list[i] - self.y_shift_list[i - 1],
                )
        average_motion_per_frame = total_motion / len(self.x_shift_list)

        # Extract results for ispyb
        fig = px.scatter(x=self.x_shift_list, y=self.y_shift_list)
        drift_plot_name = str(Path(mc_params.movie).stem) + "_drift_plot.json"
        plot_path = Path(mc_params.mrc_out).parent / drift_plot_name
        snapshot_path = Path(mc_params.mrc_out).with_suffix(".jpeg")
        fig.write_json(plot_path)

        # Forward results to ISPyB
        ispyb_parameters = {
            "ispyb_command": "buffer",
            "buffer_command": {"ispyb_command": "insert_motion_correction"},
            "buffer_store": mc_params.mc_uuid,
            "first_frame": 1,
            "last_frame": len(self.x_shift_list),
            "total_motion": total_motion,
            "average_motion_per_frame": average_motion_per_frame,
            "drift_plot_full_path": str(plot_path),
            "micrograph_snapshot_full_path": str(snapshot_path),
            "micrograph_full_path": str(mc_params.mrc_out),
            "patches_used_x": mc_params.patch_size["x"],
            "patches_used_y": mc_params.patch_size["y"],
            "dose_per_frame": mc_params.fm_dose,
        }
        self.log.info(f"Sending to ispyb {ispyb_parameters}")
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="ispyb_connector",
                message={
                    "parameters": ispyb_parameters,
                    "content": {"dummy": "dummy"},
                },
            )
        else:
            rw.send_to("ispyb_connector", ispyb_parameters)

        # If this is SPA, determine and set up the next jobs
        if mc_params.experiment_type == "spa":
            # Set up icebreaker if requested, then ctffind
            if mc_params.relion_options.do_icebreaker_jobs:
                # Three IceBreaker jobs: CtfFind job is MC+4
                ctf_job_number = 6

                # Both IceBreaker micrographs and flattening inherit from motioncorr
                self.log.info(
                    f"Sending to IceBreaker micrograph analysis: {mc_params.mrc_out}"
                )
                icebreaker_job003_params = {
                    "icebreaker_type": "micrographs",
                    "input_micrographs": mc_params.mrc_out,
                    "output_path": re.sub(
                        "MotionCorr/job002/.+",
                        "IceBreaker/job003/",
                        mc_params.mrc_out,
                    ),
                    "mc_uuid": mc_params.mc_uuid,
                    "relion_options": dict(mc_params.relion_options),
                    "total_motion": total_motion,
                    "early_motion": early_motion,
                    "late_motion": late_motion,
                }
                if isinstance(rw, MockRW):
                    rw.transport.send(
                        destination="icebreaker",
                        message={
                            "parameters": icebreaker_job003_params,
                            "content": "dummy",
                        },
                    )
                else:
                    rw.send_to("icebreaker", icebreaker_job003_params)

                self.log.info(
                    f"Sending to IceBreaker contract enhancement: {mc_params.mrc_out}"
                )
                icebreaker_job004_params = {
                    "icebreaker_type": "enhancecontrast",
                    "input_micrographs": mc_params.mrc_out,
                    "output_path": re.sub(
                        "MotionCorr/job002/.+",
                        "IceBreaker/job004/",
                        mc_params.mrc_out,
                    ),
                    "mc_uuid": mc_params.mc_uuid,
                    "relion_options": dict(mc_params.relion_options),
                    "total_motion": total_motion,
                    "early_motion": early_motion,
                    "late_motion": late_motion,
                }
                if isinstance(rw, MockRW):
                    rw.transport.send(
                        destination="icebreaker",
                        message={
                            "parameters": icebreaker_job004_params,
                            "content": "dummy",
                        },
                    )
                else:
                    rw.send_to("icebreaker", icebreaker_job004_params)

            else:
                # No IceBreaker jobs: CtfFind job is MC+1
                ctf_job_number = 3
            mc_params.ctf["output_image"] = str(
                Path(
                    mc_params.mrc_out.replace(
                        "MotionCorr/job002", f"CtfFind/job{ctf_job_number:03}"
                    )
                ).with_suffix(".ctf")
            )
            mc_params.ctf["relion_options"] = dict(mc_params.relion_options)
            mc_params.ctf["amplitude_contrast"] = mc_params.relion_options.ampl_contrast

        # Forward results to ctffind (in both SPA and tomography)
        self.log.info(f"Sending to ctf: {mc_params.mrc_out}")
        mc_params.ctf["experiment_type"] = mc_params.experiment_type
        mc_params.ctf["input_image"] = mc_params.mrc_out
        mc_params.ctf["mc_uuid"] = mc_params.mc_uuid
        mc_params.ctf["picker_uuid"] = mc_params.picker_uuid
        mc_params.ctf["pix_size"] = mc_params.pix_size
        if isinstance(rw, MockRW):
            rw.transport.send(  # type: ignore
                destination="ctffind",
                message={"parameters": mc_params.ctf, "content": "dummy"},
            )
        else:
            rw.send_to("ctffind", mc_params.ctf)

        # Forward results to images service
        self.log.info(f"Sending to images service {mc_params.mrc_out}")
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="images",
                message={
                    "image_command": "mrc_to_jpeg",
                    "file": mc_params.mrc_out,
                },
            )
        else:
            rw.send_to(
                "images",
                {
                    "image_command": "mrc_to_jpeg",
                    "file": mc_params.mrc_out,
                },
            )

        # If this is SPA, send the results to be processed by the node creator
        if mc_params.experiment_type == "spa":
            # As this is the entry point we need to import the file to the project
            self.log.info("Sending relion.import.movies to node creator")
            project_dir = Path(
                re.search(".+/job[0-9]{3}/", mc_params.mrc_out)[0]
            ).parent.parent
            import_movie = (
                project_dir
                / "Import/job001"
                / Path(mc_params.mrc_out)
                .relative_to(project_dir / "MotionCorr/job002")
                .parent
                / Path(mc_params.movie).name
            )
            if not import_movie.parent.is_dir():
                import_movie.parent.mkdir(parents=True)
            import_movie.unlink(missing_ok=True)
            import_movie.symlink_to(mc_params.movie)
            import_parameters = {
                "job_type": "relion.import.movies",
                "input_file": str(mc_params.movie),
                "output_file": str(import_movie),
                "relion_options": dict(mc_params.relion_options),
                "command": "",
                "stdout": "",
                "stderr": "",
            }
            if isinstance(rw, MockRW):
                rw.transport.send(
                    destination="node_creator",
                    message={"parameters": import_parameters, "content": "dummy"},
                )
            else:
                rw.send_to("node_creator", import_parameters)

            # Then register the motion correction job with the node creator
            self.log.info(f"Sending {self.job_type} to node creator")
            node_creator_parameters = {
                "job_type": self.job_type,
                "input_file": str(import_movie),
                "output_file": mc_params.mrc_out,
                "relion_options": dict(mc_params.relion_options),
                "command": " ".join(command),
                "stdout": result.stdout.decode("utf8", "replace"),
                "stderr": result.stderr.decode("utf8", "replace"),
                "results": {
                    "total_motion": total_motion,
                    "early_motion": early_motion,
                    "late_motion": late_motion,
                },
            }
            if isinstance(rw, MockRW):
                rw.transport.send(
                    destination="node_creator",
                    message={"parameters": node_creator_parameters, "content": "dummy"},
                )
            else:
                rw.send_to("node_creator", node_creator_parameters)

        # Register completion with Murfey if this is tomography
        if mc_params.experiment_type == "tomography":
            self.log.info("Sending to Murfey")
            if isinstance(rw, MockRW):
                rw.transport.send(
                    "murfey_feedback",
                    {
                        "register": "motion_corrected",
                        "movie": mc_params.movie,
                        "mrc_out": mc_params.mrc_out,
                        "movie_id": mc_params.movie_id,
                    },
                )
            else:
                rw.send_to(
                    "murfey_feedback",
                    {
                        "register": "motion_corrected",
                        "movie": mc_params.movie,
                        "mrc_out": mc_params.mrc_out,
                        "movie_id": mc_params.movie_id,
                    },
                )

        self.log.info(f"Done {self.job_type} for {mc_params.movie}.")
        rw.transport.ack(header)
        self.x_shift_list = []
        self.y_shift_list = []



from __future__ import annotations

import errno
import logging
import os
import re
import subprocess
import time
from typing import Any, Callable, Dict, NamedTuple, Protocol

import PIL.Image
import workflows.recipe
from importlib_metadata import entry_points
from workflows.services.common_service import CommonService

logger = logging.getLogger("relion.zocalo.images")


class _CallableParameter(Protocol):
    def __call__(self, key: str, default: Any = ...) -> Any:
        ...


class PluginInterface(NamedTuple):
    rw: workflows.recipe.wrapper.RecipeWrapper
    parameters: _CallableParameter
    message: Dict[str, Any]


class Images(CommonService):
    """
    A service that generates images and thumbnails.
    Plugin functions can be registered under the entry point
    'zocalo.services.images.plugins'. The contract is that a plugin function
    takes a single argument of type PluginInterface, and returns a truthy value
    to acknowledge success, and a falsy value to reject the related message.
    If a falsy value is returned that is not False then, additionally, an error
    is logged.
    Functions may choose to return a list of files that were generated, but
    this is optional at this time.
    """

    # Human readable service name
    _service_name = "Images"

    # Logger name
    _logger_name = "relion.zocalo.images"

    # Dictionary to contain functions from plugins
    image_functions: dict[str, Callable] = {}

    def initializing(self):
        """Subscribe to a queue. Received messages must be acknowledged."""
        self.log.info("Image service starting")
        self.image_functions.update(
            {
                e.name: e.load()
                for e in entry_points(group="zocalo.services.images.plugins")
            }
        )
        workflows.recipe.wrap_subscribe(
            self._transport,
            "images",
            self.image_call,
            acknowledgement=True,
            log_extender=self.extend_log,
        )

    def image_call(self, rw, header, message):
        """Pass incoming message to the relevant plugin function."""

        def parameters(key: str, default=None):
            if isinstance(message, dict) and message.get(key):
                return message[key]
            return rw.recipe_step.get("parameters", {}).get(key, default)

        command = parameters("image_command")
        if command not in self.image_functions:
            self.log.error(f"Unknown command: {command!r}")
            rw.transport.nack(header)
            return

        start = time.perf_counter()
        try:
            result = self.image_functions[command](
                PluginInterface(rw, parameters, message)
            )
        except (PermissionError, FileNotFoundError) as e:
            self.log.error(f"Command {command!r} raised {e}", exc_info=True)
            rw.transport.nack(header)
            return
        runtime = time.perf_counter() - start

        if result:
            self.log.info(f"Command {command!r} completed in {runtime:.1f} seconds")
            rw.transport.ack(header)
        elif result is False:
            # The assumption here is that if a function returns explicit
            # 'False' then it has already taken care of logging, so we
            # don't need yet another log record.
            rw.transport.nack(header)
        else:
            self.log.error(
                f"Command {command!r} returned {result!r} after {runtime:.1f} seconds"
            )
            rw.transport.nack(header)


def diffraction(plugin: PluginInterface):
    """Take a diffraction data file and transform it into JPEGs."""
    filename = plugin.parameters("file")

    imageset_index = 1
    if not filename:
        # 'file' is a filename
        # 'input' is a xia2-type string, may need to remove :x:x suffix
        filename = plugin.parameters("input")
        if ":" in filename:
            filename, imageset_index = filename.split(":")[0:2]

    if not filename or filename == "None":
        logger.debug("Skipping diffraction JPG generation: filename not specified")
        return False
    if not os.path.exists(filename):
        logger.error("File %s not found", filename)
        return False
    sizex = plugin.parameters("size-x", default=400)
    sizey = plugin.parameters("size-y", default=192)
    output = plugin.parameters("output")
    if not output:
        # split off extension
        output = filename[: filename.rindex(".")]
        # deduct image filename
        output = re.sub(
            r"(/[a-z]{2}[0-9]{4,}-[0-9]+/)", r"\g<0>jpegs/", output, count=1
        )
        output = output + ".jpeg"
        # create directory for image if necessary
        try:
            os.makedirs(os.path.dirname(output))
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
    output_small = output[: output.rindex(".")] + ".thumb.jpeg"

    start = time.perf_counter()
    result = subprocess.run(
        [
            "dials.export_bitmaps",
            filename,
            "imageset_index=%s" % imageset_index,
            "output.format=jpeg",
            "quality=95",
            "binning=4",
            "brightness=60",
            'output.file="%s"' % output,
        ]
    )
    export = time.perf_counter()
    if result.returncode:
        logger.error(
            f"Export of {filename} failed with exitcode {result.returncode}:\n"
            + result.stderr.decode("utf8", "replace")
        )
        return False
    if not os.path.exists(output):
        logger.error("Output file %s not found", output)
        return False
    with PIL.Image.open(output) as fh:
        fh.thumbnail((sizex, sizey))
        fh.save(output_small)
    done = time.perf_counter()

    logger.info(
        "Created thumbnail %s -> %s (%.1f sec) -> %s (%.1f sec)",
        filename,
        output,
        export - start,
        output_small,
        done - export,
    )
    return [output, output_small]


def thumbnail(plugin: PluginInterface):
    """Take a single file and create a smaller version of the same file."""
    filename = plugin.parameters("file")
    if not filename or filename == "None":
        logger.debug("Skipping thumbnail generation: filename not specified")
        return False
    if not os.path.exists(filename):
        logger.error("File %s not found", filename)
        return False
    sizex = plugin.parameters("size-x", default=400)
    sizey = plugin.parameters("size-y", default=192)
    output = plugin.parameters("output")
    if not output:
        # If not set add a 't' in front of the last '.' in the filename
        output = (
            filename[: filename.rindex(".")] + "t" + filename[filename.rindex(".") :]
        )

    start = time.perf_counter()
    with PIL.Image.open(filename) as fh:
        fh.thumbnail((sizex, sizey))
        fh.save(output)
    timing = time.perf_counter() - start

    logger.info("Created thumbnail %s -> %s in %.1f seconds", filename, output, timing)
    return [output]



from __future__ import annotations

import datetime
import errno
import json
import logging
import math
import os
import pathlib
import subprocess
from pprint import pformat
from typing import Optional

import pydantic
import requests
import workflows.recipe
import zocalo.configuration
from importlib_metadata import entry_points
from workflows.services.common_service import CommonService
from zocalo.util import slurm


class JobSubmissionParameters(pydantic.BaseModel):
    scheduler: str = "grid_engine"
    cluster: Optional[str]
    partition: Optional[str]
    job_name: Optional[str]  #
    environment: Optional[dict[str, str]] = None
    cpus_per_task: Optional[int] = None
    tasks: Optional[int] = None  # slurm only
    nodes: Optional[int]  # slurm only
    memory_per_node: Optional[int] = None  # slurm only
    gpus_per_node: Optional[str] = None  # slurm only
    min_memory_per_cpu: Optional[int] = pydantic.Field(
        None, description="Minimum real memory per cpu (MB)"
    )
    max_memory_per_cpu: Optional[int] = pydantic.Field(
        None, description="Maximum real memory per cpu (MB)"
    )  # HTCondor: maximum memory allocated for job
    max_disk_per_cpu: Optional[int] = pydantic.Field(
        None, description="Maximum disk space per cpu (MB)"
    )  # HTCondor: maximum disk space allocated for job
    time_limit: Optional[datetime.timedelta] = None
    gpus: Optional[int] = None
    exclusive: bool = False
    account: Optional[str]  # account in slurm terminology
    commands: str | list[str]
    qos: Optional[str]
    queue: Optional[str]  # legacy for grid engine
    transfer_input_files: Optional[list[str]]  # HTCondor: list of inputs to transfer
    transfer_output_files: Optional[list[str]]  # HTCondor: list of outputs to transfer


def format_timedelta_to_HHMMSS(td: datetime.timedelta) -> str:
    td_in_seconds = td.total_seconds()
    hours, remainder = divmod(td_in_seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    return f"{hours:02.0f}:{minutes:02.0f}:{seconds:02.0f}"


def submit_to_grid_engine(
    params: JobSubmissionParameters,
    working_directory: pathlib.Path,
    logger: logging.Logger,
    **kwargs,
) -> int | None:
    # validate
    assert params.cluster is not None

    submission_params: list = []

    if params.job_name:
        submission_params.extend(["-N", params.job_name])
    if params.cpus_per_task:
        submission_params.extend(["-pe", "smp", str(params.cpus_per_task)])
    if params.min_memory_per_cpu:
        submission_params.extend(["-l", f"mfree={params.min_memory_per_cpu}M"])
    if params.time_limit:
        HHMMSS = format_timedelta_to_HHMMSS(params.time_limit)
        submission_params.extend(["-l", f"h_rt={HHMMSS}"])
    if params.exclusive:
        submission_params.extend(["-l", "exclusive"])
    if params.gpus:
        submission_params.extend(["-l", f"gpus={params.gpus}"])
    if params.account:
        submission_params.extend(["-P", params.account])
    if params.queue:
        submission_params.extend(["-q", params.queue])

    commands = params.commands
    if not isinstance(commands, str):
        commands = " ".join(commands)

    submission = [
        ". /etc/profile.d/modules.sh",
        "module load global/" + params.cluster,
        # thanks to Modules 3.2 weirdness qsub may now be a function
        # calling the real qsub command, but eating up its parameters.
        "unset -f qsub",
        f"qsub {' '.join(submission_params)} << EOF",
        "#!/bin/bash",
        ". /etc/profile.d/modules.sh",
        "cd " + os.fspath(working_directory),
        commands,
        "EOF",
    ]
    logger.debug(
        f"Cluster ({params.cluster}) submission parameters: {submission_params}"
    )
    logger.debug(f"Commands: {commands}")
    logger.debug(f"Working directory: {working_directory}")
    result = subprocess.run(
        ["/bin/bash"],
        input="\n".join(submission).encode("latin1"),
        cwd=working_directory,
        capture_output=True,
    )
    if result.returncode:
        logger.error(
            "Could not submit cluster job:\n%s\n%s",
            result.stdout.decode("latin1"),
            result.stderr.decode("latin1"),
        )
        return None
    assert b"has been submitted" in result.stdout
    jobnumber = result.stdout.split()[2].decode("latin1")
    return int(jobnumber)


def submit_to_slurm(
    params: JobSubmissionParameters,
    working_directory: pathlib.Path,
    logger: logging.Logger,
    zc: zocalo.configuration,
) -> int | None:
    api = slurm.SlurmRestApi.from_zocalo_configuration(zc)

    script = params.commands
    if not isinstance(script, str):
        script = "\n".join(script)
    script = f"#!/bin/bash\n. /etc/profile.d/modules.sh\n{script}"

    logger.debug(f"Submitting script to Slurm:\n{script}")
    if params.time_limit:
        time_limit_minutes = math.ceil(params.time_limit.total_seconds() / 60)
    else:
        time_limit_minutes = None
    job_submission = slurm.models.JobSubmission(
        script=script,
        job=slurm.models.JobProperties(
            partition=params.partition,
            name=params.job_name,
            cpus_per_task=params.cpus_per_task,
            tasks=params.tasks,
            nodes=[params.nodes, params.nodes] if params.nodes else params.nodes,
            gpus_per_node=params.gpus_per_node,
            memory_per_node=params.memory_per_node,
            environment=os.environ
            if params.environment is None
            else params.environment,
            memory_per_cpu=params.min_memory_per_cpu,
            time_limit=time_limit_minutes,
            gpus=params.gpus,
            # exclusive=params.exclusive,
            account=params.account,
            current_working_directory=os.fspath(working_directory),
            qos=params.qos,
        ),
    )
    try:
        response = api.submit_job(job_submission)
    except requests.HTTPError as e:
        logger.error(f"Failed Slurm job submission: {e}\n" f"{e.response.text}")
        return None
    if response.errors:
        error_message = "\n".join(f"{e.errno}: {e.error}" for e in response.errors)
        logger.error(f"Failed Slurm job submission: {error_message}")
        return None
    return response.job_id


def submit_to_htcondor(
    params: JobSubmissionParameters,
    working_directory: pathlib.Path,
    logger: logging.Logger,
    **kwargs,
) -> int | None:
    current_wd = os.getcwd()

    singularity_environment = "SINGULARITY_CACHEDIR=/tmp/singularity SINGULARITY_LOCALCACHEDIR=/tmp/singularity SINGULARITY_TMPDIR=/tmp/singularity"

    commands = params.commands
    if not isinstance(commands, str):
        commands = "\n".join(commands)
    cluster_exec, cluster_args = commands.split("\n", 1)
    logger.info(f"{cluster_exec} {cluster_args}")
    htcondor_submit = {
        "executable": cluster_exec,
        "arguments": cluster_args,
        "universe": "vanilla",
        "environment": singularity_environment
        if params.environment is None
        else " ".join(f"{k}={v}" for k, v in params.environment.items()),
        "should_transfer_files": "YES",
        "when_to_transfer_output": "ON_EXIT_OR_EVICT",
        "output": f"{params.job_name}.condor.out",
        "error": f"{params.job_name}.condor.err",
        "log": f"{params.job_name}.condor.log",
        "on_exit_hold": False,
        "on_exit_remove": True,
    }
    if params.cpus_per_task:
        htcondor_submit.update({"request_cpus": str(params.cpus_per_task)})
        if params.max_memory_per_cpu:
            htcondor_submit.update(
                {
                    "request_memory": f"{str(params.max_memory_per_cpu * params.cpus_per_task)}MB"
                }
            )
        if params.max_disk_per_cpu:
            htcondor_submit.update(
                {
                    "request_disk": f"{str(params.max_disk_per_cpu * params.cpus_per_task)}MB"
                }
            )
    if params.transfer_input_files:
        htcondor_submit.update(
            {"transfer_input_files": ",".join(params.transfer_input_files)}
        )
    if params.transfer_output_files:
        htcondor_submit.update(
            {"transfer_output_files": ",".join(params.transfer_output_files)}
        )

    try:
        import htcondor

        coll = htcondor.Collector(htcondor.param["COLLECTOR_HOST"])
        schedd_ad = coll.locate(htcondor.DaemonTypes.Schedd)
        schedd = htcondor.Schedd(schedd_ad)
        logger.debug(f"Address of the Schedd is: {str(schedd_ad['MyAddress'])}")
        os.chdir(working_directory)
        htcondor_job = htcondor.Submit(htcondor_submit)

        with schedd.transaction() as txn:
            jobnumber = htcondor_job.queue(txn, count=1)
    except Exception:
        logger.exception(
            f"Could not submit HTCondor job:\n{pformat(htcondor_submit)}",
        )
        return None
    finally:
        os.chdir(current_wd)
    return jobnumber


class ClusterSubmission(CommonService):
    """A service to interface zocalo with functions to start new
    jobs on the clusters."""

    # Human readable service name
    _service_name = "Cluster submission service"

    # Logger name
    _logger_name = "relion.zocalo.cluster"

    def initializing(self):
        """Subscribe to the cluster submission queue.
        Received messages must be acknowledged."""
        self.log.info("Cluster submission service starting")

        if not self.environment_is_valid():
            self._request_termination()
            return

        self.schedulers = {
            f.name: f.load()
            for f in entry_points(group="zocalo.services.cluster.schedulers")
        }
        self.log.debug(f"Supported schedulers: {', '.join(self.schedulers.keys())}")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "cluster.submission",
            self.run_submit_job,
            acknowledgement=True,
            log_extender=self.extend_log,
        )

    def environment_is_valid(self):
        """Check that the cluster submission environment is sane. Specifically, that
        there is no ~/.sge_request file interfering with cluster job submissions.
        """
        sge_file = pathlib.Path("~").expanduser() / ".sge_request"
        if sge_file.exists():
            contents = sge_file.read_bytes().strip()
            if contents:
                self.log.error(
                    "Rejecting service initialisation: file %s is not empty. "
                    "This may interfere with service operation. ",
                    str(sge_file),
                )
                return False

            self.log.info(
                "Note: empty file %s found during service startup", str(sge_file)
            )
        return True

    @staticmethod
    def _recursive_mkdir(path):
        try:
            os.makedirs(path)
        except OSError as exc:
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                raise

    def run_submit_job(self, rw, header, message):
        """Submit cluster job according to message."""

        parameters = rw.recipe_step["parameters"]
        cluster_params = JobSubmissionParameters(**parameters.get("cluster", {}))

        if not isinstance(cluster_params.commands, str):
            cluster_params.commands = "\n".join(cluster_params.commands)

        if "recipefile" in parameters:
            recipefile = parameters["recipefile"]
            try:
                self._recursive_mkdir(os.path.dirname(recipefile))
            except OSError as e:
                if e.errno == errno.ENOENT:
                    self.log.error(
                        "Error in underlying filesystem: %s", str(e), exc_info=True
                    )
                    self._transport.nack(header)
                    return
                raise
            self.log.debug("Writing recipe to %s", recipefile)
            cluster_params.commands = cluster_params.commands.replace(
                "$RECIPEFILE", recipefile
            )
            with open(recipefile, "w") as fh:
                fh.write(rw.recipe.pretty())
        if "recipeenvironment" in parameters:
            recipeenvironment = parameters["recipeenvironment"]
            try:
                self._recursive_mkdir(os.path.dirname(recipeenvironment))
            except OSError as e:
                if e.errno == errno.ENOENT:
                    self.log.error(
                        "Error in underlying filesystem: %s", str(e), exc_info=True
                    )
                    self._transport.nack(header)
                    return
                raise
            self.log.debug("Writing recipe environment to %s", recipeenvironment)
            cluster_params.commands = cluster_params.commands.replace(
                "$RECIPEENV", recipeenvironment
            )
            with open(recipeenvironment, "w") as fh:
                json.dump(
                    rw.environment, fh, sort_keys=True, indent=2, separators=(",", ": ")
                )
        if "recipewrapper" in parameters:
            recipewrapper = parameters["recipewrapper"]
            try:
                self._recursive_mkdir(os.path.dirname(recipewrapper))
            except OSError as e:
                if e.errno == errno.ENOENT:
                    self.log.error(
                        "Error in underlying filesystem: %s", str(e), exc_info=True
                    )
                else:
                    self.log.error(
                        "Could not create working directory: %s", str(e), exc_info=True
                    )
                self._transport.nack(header)
                return
            self.log.debug("Storing serialized recipe wrapper in %s", recipewrapper)
            cluster_params.commands = cluster_params.commands.replace(
                "$RECIPEWRAP", recipewrapper
            )
            with open(recipewrapper, "w") as fh:
                json.dump(
                    {
                        "recipe": rw.recipe.recipe,
                        "recipe-pointer": rw.recipe_pointer,
                        "environment": rw.environment,
                        "recipe-path": rw.recipe_path,
                        "payload": rw.payload,
                    },
                    fh,
                    indent=2,
                    separators=(",", ": "),
                )

        if "workingdir" not in parameters or not parameters["workingdir"].startswith(
            "/"
        ):
            self.log.error(
                "No absolute working directory specified. Will not run cluster job"
            )
            self._transport.nack(header)
            return
        working_directory = pathlib.Path(parameters["workingdir"])
        try:
            working_directory.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            self.log.error(
                "Could not create working directory: %s", str(e), exc_info=True
            )
            self._transport.nack(header)
            return

        submit_to_scheduler = self.schedulers.get(cluster_params.scheduler)

        jobnumber = submit_to_scheduler(
            cluster_params, working_directory, self.log, zc=self.config
        )
        if not jobnumber:
            self._transport.nack(header)
            return

        # Conditionally acknowledge receipt of the message
        txn = self._transport.transaction_begin(subscription_id=header["subscription"])
        self._transport.ack(header, transaction=txn)

        # Send results onwards
        rw.set_default_channel("job_submitted")
        rw.send({"jobid": jobnumber}, transaction=txn)

        # Commit transaction
        self._transport.transaction_commit(txn)
        self.log.info(f"Submitted job {jobnumber} to {cluster_params.cluster}")



from __future__ import annotations

import json
import re
import shutil
import subprocess
from pathlib import Path

import numpy as np
import workflows.recipe
from gemmi import cif
from pydantic import BaseModel, Field, ValidationError
from workflows.services.common_service import CommonService

from relion.zocalo.spa_relion_service_options import RelionServiceOptions


class SelectClassesParameters(BaseModel):
    input_file: str = Field(..., min_length=1)
    combine_star_job_number: int
    particles_file: str = "particles.star"
    classes_file: str = "class_averages.star"
    python_exe: str = "python"
    min_score: float = 0
    min_particles: int = 500
    class3d_batch_size: int = 50000
    class3d_max_size: int = 200000
    program_id: int
    session_id: int
    class_uuids: str
    relion_options: RelionServiceOptions


class SelectClasses(CommonService):
    """
    A service for running Relion autoselection on 2D classes
    """

    # Human readable service name
    _service_name = "SelectClasses"

    # Logger name
    _logger_name = "relion.zocalo.select_classes"

    # Job name
    job_type = "relion.select.class2dauto"

    # Values to extract for ISPyB
    previous_total_count = 0
    total_count = 0

    # Values for ISPyB lookups
    class_uuids_dict: dict = {}
    class_uuids_keys: list = []

    def initializing(self):
        """Subscribe to a queue. Received messages must be acknowledged."""
        self.log.info("Select particles service starting")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "select_classes",
            self.select_classes,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    def parse_combiner_output(self, combiner_stdout: str):
        """
        Read the output logs of the star file combination
        """
        for line in combiner_stdout.split("\n"):
            if line.startswith("Adding") and "particles_all.star" in line:
                line_split = line.split()
                self.previous_total_count = int(line_split[3])

            if line.startswith("Combined"):
                line_split = line.split()
                self.total_count = int(line_split[6])

    def select_classes(self, rw, header: dict, message: dict):
        class MockRW:
            def dummy(self, *args, **kwargs):
                pass

        if not rw:
            print(
                "Incoming message is not a recipe message. Simple messages can be valid"
            )
            if (
                not isinstance(message, dict)
                or not message.get("parameters")
                or not message.get("content")
            ):
                self.log.error("Rejected invalid simple message")
                self._transport.nack(header)
                return
            self.log.debug("Received a simple message")

            # Create a wrapper-like object that can be passed to functions
            # as if a recipe wrapper was present.
            rw = MockRW()
            rw.transport = self._transport
            rw.recipe_step = {"parameters": message["parameters"]}
            rw.environment = {"has_recipe_wrapper": False}
            rw.set_default_channel = rw.dummy
            rw.send = rw.dummy
            message = message["content"]

        try:
            if isinstance(message, dict):
                autoselect_params = SelectClassesParameters(
                    **{**rw.recipe_step.get("parameters", {}), **message}
                )
            else:
                autoselect_params = SelectClassesParameters(
                    **{**rw.recipe_step.get("parameters", {})}
                )
        except (ValidationError, TypeError) as e:
            self.log.warning(
                f"Selection parameter validation failed for message: {message} "
                f"and recipe parameters: {rw.recipe_step.get('parameters', {})} "
                f"with exception: {e}"
            )
            rw.transport.nack(header)
            return

        self.log.info(f"Inputs: {autoselect_params.input_file}")

        class2d_job_dir = Path(
            re.search(".+/job[0-9]{3}/", autoselect_params.input_file)[0]
        )
        project_dir = class2d_job_dir.parent.parent

        select_job_num = (
            int(re.search("/job[0-9]{3}", str(class2d_job_dir))[0][4:7]) + 2
        )
        select_dir = project_dir / f"Select/job{select_job_num:03}"
        select_dir.mkdir(parents=True, exist_ok=True)

        autoselect_flags = {
            "particles_file": "--fn_sel_parts",
            "classes_file": "--fn_sel_classavgs",
            "python_exe": "--python",
            "min_particles": "--select_min_nr_particles",
        }
        # Create the class selection command
        autoselect_command = [
            "relion_class_ranker",
            "--opt",
            autoselect_params.input_file,
            "--o",
            f"{select_dir.relative_to(project_dir)}/",
            "--auto_select",
            "--fn_root",
            "rank",
            "--do_granularity_features",
        ]
        for k, v in autoselect_params.dict().items():
            if v and (k in autoselect_flags):
                autoselect_command.extend((autoselect_flags[k], str(v)))
        autoselect_command.extend(
            ("--pipeline_control", f"{select_dir.relative_to(project_dir)}/")
        )

        if not autoselect_params.min_score:
            autoselect_command.extend(("--min_score", "0.0"))
        else:
            autoselect_command.extend(("--min_score", str(autoselect_params.min_score)))

        # Run the class selection
        result = subprocess.run(
            autoselect_command, cwd=str(project_dir), capture_output=True
        )

        if not autoselect_params.min_score and not result.returncode:
            # If a minimum score isn't given, then work it out and rerun the job
            star_doc = cif.read_file(str(select_dir / "rank_model.star"))
            star_block = star_doc["model_classes"]
            class_scores = np.array(star_block.find_loop("_rlnClassScore"), dtype=float)
            quantile_threshold = np.quantile(
                class_scores,
                float(
                    autoselect_params.relion_options.class2d_fraction_of_classes_to_remove
                ),
            )

            self.log.info(f"Sending new threshold {quantile_threshold} to Murfey")
            murfey_params = {
                "register": "save_class_selection_score",
                "class_selection_score": quantile_threshold,
                "program_id": autoselect_params.program_id,
                "session_id": autoselect_params.session_id,
            }
            if isinstance(rw, MockRW):
                rw.transport.send(destination="murfey_feedback", message=murfey_params)
            else:
                rw.send_to("murfey_feedback", murfey_params)

            self.log.info(
                f"Re-running class selection with new threshold {quantile_threshold}"
            )
            autoselect_command[-1] = str(quantile_threshold)

            # Re-run the class selection
            result = subprocess.run(
                autoselect_command, cwd=str(project_dir), capture_output=True
            )

        # Send class selection job to node creator
        self.log.info(f"Sending {self.job_type} to node creator")
        autoselect_node_creator_params = {
            "job_type": self.job_type,
            "input_file": autoselect_params.input_file,
            "output_file": str(select_dir / autoselect_params.particles_file),
            "relion_options": dict(autoselect_params.relion_options),
            "command": " ".join(autoselect_command),
            "stdout": result.stdout.decode("utf8", "replace"),
            "stderr": result.stderr.decode("utf8", "replace"),
        }
        if result.returncode:
            autoselect_node_creator_params["success"] = False
        else:
            autoselect_node_creator_params["success"] = True
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="node_creator",
                message={
                    "parameters": autoselect_node_creator_params,
                    "content": "dummy",
                },
            )
        else:
            rw.send_to("node_creator", autoselect_node_creator_params)

        # End here if the command failed
        if result.returncode:
            self.log.error(
                f"2D autoselection failed with exitcode {result.returncode}:\n"
                + result.stderr.decode("utf8", "replace")
            )
            rw.transport.nack(header)
            return

        # Find which classes were picked
        classes_block = cif.read_file(f"{select_dir}/class_averages.star").sole_block()
        selected_classes = list(classes_block.find_loop("_rlnReferenceImage"))

        # Class ids get fed in as a string, need to convert these to a dictionary
        self.class_uuids_dict = json.loads(
            autoselect_params.class_uuids.replace("'", '"')
        )
        self.class_uuids_keys = list(self.class_uuids_dict.keys())

        # Send the picked classes to ispyb
        ispyb_command_list = []
        for picked_class in selected_classes:
            class_id = int(picked_class.split("@")[0])
            ispyb_command_list.append(
                {
                    "ispyb_command": "buffer",
                    "buffer_lookup": {
                        "particle_classification_id": self.class_uuids_dict[
                            self.class_uuids_keys[class_id - 1]
                        ],
                    },
                    "buffer_command": {
                        "ispyb_command": "insert_particle_classification"
                    },
                    "selected": 1,
                }
            )
        ispyb_parameters = {
            "ispyb_command": "multipart_message",
            "ispyb_command_list": ispyb_command_list,
        }
        self.log.info(f"Sending to ispyb {ispyb_parameters}")
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="ispyb_connector",
                message={
                    "parameters": ispyb_parameters,
                    "content": {"dummy": "dummy"},
                },
            )
        else:
            rw.send_to("ispyb_connector", ispyb_parameters)

        # Run the combine star files job to combine the files into particles_all.star
        self.log.info("Running star file combination and splitting")
        combine_star_command = [
            "combine_star_files.py",
            str(select_dir / autoselect_params.particles_file),
        ]

        combine_star_dir = Path(
            project_dir / f"Select/job{autoselect_params.combine_star_job_number:03}"
        )
        if (combine_star_dir / "particles_all.star").exists():
            combine_star_command.append(str(combine_star_dir / "particles_all.star"))
        else:
            combine_star_dir.mkdir(parents=True, exist_ok=True)
            Path(project_dir / "Select/Star_combination").symlink_to(combine_star_dir)
            self.previous_total_count = 0
        combine_star_command.extend(("--output_dir", str(combine_star_dir)))

        result = subprocess.run(
            combine_star_command, cwd=str(project_dir), capture_output=True
        )
        self.parse_combiner_output(result.stdout.decode("utf8", "replace"))

        # Send combination job to node creator
        self.log.info("Sending combine_star_files_job (combine) to node creator")
        combine_node_creator_params = {
            "job_type": "combine_star_files_job",
            "input_file": f"{select_dir}/{autoselect_params.particles_file}",
            "output_file": f"{combine_star_dir}/particles_all.star",
            "relion_options": dict(autoselect_params.relion_options),
            "command": " ".join(combine_star_command),
            "stdout": result.stdout.decode("utf8", "replace"),
            "stderr": result.stderr.decode("utf8", "replace"),
        }
        if result.returncode:
            combine_node_creator_params["success"] = False
        else:
            combine_node_creator_params["success"] = True
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="node_creator",
                message={"parameters": combine_node_creator_params, "content": "dummy"},
            )
        else:
            rw.send_to("node_creator", combine_node_creator_params)

        # End here if the command failed
        if result.returncode:
            self.log.error(
                f"Star file combination failed with exitcode {result.returncode}:\n"
                + result.stderr.decode("utf8", "replace")
            )
            rw.transport.nack(header)
            return

        # Determine the next split size to use and whether to run 3D classification
        send_to_3d_classification = False
        if self.previous_total_count == 0:
            # First run of this job, use class3d_max_size
            next_batch_size = autoselect_params.class3d_batch_size
            if self.total_count > autoselect_params.class3d_batch_size:
                # Do 3D classification if there are more particles than the batch size
                send_to_3d_classification = True
        elif self.previous_total_count >= autoselect_params.class3d_max_size:
            # Iterations beyond those where 3D classification is run
            next_batch_size = autoselect_params.class3d_max_size
        else:
            # Re-runs with fewer particles than the maximum
            previous_batch_multiple = (
                self.previous_total_count // autoselect_params.class3d_batch_size
            )
            new_batch_multiple = (
                self.total_count // autoselect_params.class3d_batch_size
            )
            if new_batch_multiple > previous_batch_multiple:
                # Do 3D classification if a batch threshold has been crossed
                send_to_3d_classification = True
                # Set the batch size from the total count, but do not exceed the maximum
                next_batch_size = (
                    new_batch_multiple * autoselect_params.class3d_batch_size
                )
                if next_batch_size > autoselect_params.class3d_max_size:
                    next_batch_size = autoselect_params.class3d_max_size
            else:
                # Otherwise just get the next threshold
                next_batch_size = (
                    previous_batch_multiple + 1
                ) * autoselect_params.class3d_batch_size

        # Run the combine star files job to split particles_all.star into batches
        split_star_command = [
            "combine_star_files.py",
            str(combine_star_dir / "particles_all.star"),
            "--output_dir",
            str(combine_star_dir),
            "--split",
            "--split_size",
            str(next_batch_size),
        ]

        result = subprocess.run(
            split_star_command, cwd=str(project_dir), capture_output=True
        )
        self.parse_combiner_output(result.stdout.decode("utf8", "replace"))

        # Send splitting job to node creator
        self.log.info("Sending combine_star_files_job (split) to node creator")
        combine_node_creator_params = {
            "job_type": "combine_star_files_job",
            "input_file": f"{select_dir}/{autoselect_params.particles_file}",
            "output_file": f"{combine_star_dir}/particles_all.star",
            "relion_options": dict(autoselect_params.relion_options),
            "command": " ".join(split_star_command),
            "stdout": result.stdout.decode("utf8", "replace"),
            "stderr": result.stderr.decode("utf8", "replace"),
        }
        if result.returncode:
            combine_node_creator_params["success"] = False
        else:
            combine_node_creator_params["success"] = True
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="node_creator",
                message={"parameters": combine_node_creator_params, "content": "dummy"},
            )
        else:
            rw.send_to("node_creator", combine_node_creator_params)

        # End here if the command failed
        if result.returncode:
            self.log.error(
                f"Star file splitting failed with exitcode {result.returncode}:\n"
                + result.stderr.decode("utf8", "replace")
            )
            rw.transport.nack(header)
            return

        # Create 3D classification jobs
        if send_to_3d_classification:
            # Only send to 3D if a new multiple of the batch threshold is crossed
            # and the count has not passed the maximum
            self.log.info("Sending to Murfey for Class3D")

            # Copy the particle batch file
            shutil.copy2(
                combine_star_dir / "particles_split1.star",
                combine_star_dir / f"particles_batch_{next_batch_size}.star",
            )

            # Tell Murfey to do Class3D
            class3d_params = {
                "particles_file": f"{combine_star_dir}/particles_batch_{next_batch_size}.star",
                "class3d_dir": f"{project_dir}/Class3D/job",
                "batch_size": next_batch_size,
            }
            murfey_3d_params = {
                "register": "run_class3d",
                "class3d_message": class3d_params,
                "program_id": autoselect_params.program_id,
                "session_id": autoselect_params.session_id,
            }
            if isinstance(rw, MockRW):
                rw.transport.send("murfey_feedback", murfey_3d_params)
            else:
                rw.send_to("murfey_feedback", murfey_3d_params)

        murfey_confirmation = {
            "register": "done_class_selection",
            "program_id": autoselect_params.program_id,
            "session_id": autoselect_params.session_id,
        }
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="murfey_feedback", message=murfey_confirmation
            )
        else:
            rw.send_to("murfey_feedback", murfey_confirmation)

        self.log.info(f"Done {self.job_type} for {autoselect_params.input_file}.")
        rw.transport.ack(header)
